unit Dpi.Common;

interface

uses
  {$IFDEF WINDOWS}Windows,{$ENDIF}
  Classes, SysUtils, LCLType, Types, LCLIntf, Graphics, Math, Dpi.Graphics;

const
  DpiControlsComponentPaletteName = 'DpiControls';

function BitBlt(DestDC: HDC; X, Y, Width, Height: Integer; SrcDC: HDC; XSrc,
    YSrc: Integer; Rop: DWORD = SRCCOPY): Boolean;
function BitBltCanvas(Dest: TCanvas; X, Y, Width, Height: Integer; Src: TCanvas;
    XSrc, YSrc: Integer; Rop: DWORD = SRCCOPY): Boolean;
function CreateRectRgn(X1, Y1, X2, Y2: Integer): HRGN;
{$IFDEF WINDOWS}
function ScrollDC(hDC: HDC; dx: longint; dy: longint; const lprcScroll: RECT;
  const lprcClip:RECT;hrgnUpdate:HRGN; lprcUpdate: LPRECT): WINBOOL; overload;
{$ENDIF}
function ScrollDC(Canvas: TCanvas; dx: Longint; dy: Longint; const lprcScroll: TRect;
  const lprcClip:TRect; hrgnUpdate: Handle; lprcUpdate: PRect): Boolean; overload;
function SetWindowPos(hWnd: HWND; hWndInsertAfter: HWND;
  X, Y, cx, cy: Integer; uFlags: UINT): Boolean;
function ScaleToNative(Value: Integer): Integer;
function ScaleToNativeDist(Base, Value: Integer): Integer;
function ScaleFromNative(Value: Integer): Integer;
function ScalePointToNative(Value: TPoint): TPoint;
function ScalePointFromNative(Value: TPoint): TPoint;
function ScaleSizeToNative(Value: TSize): TSize;
function ScaleSizeFromNative(Value: TSize): TSize;
function ScaleRectToNative(Value: TRect): TRect;
function ScaleRectFromNative(Value: TRect): TRect;
function ScaleFloatToNative(Value: Double): Double;
function ScaleFloatFromNative(Value: Double): Double;
procedure WriteLog(Text: string);
function GetSystemMetrics(nIndex: Integer): Integer;

resourcestring
  SNotImplemented = 'Not implemented';


implementation

function BitBltCanvas(Dest: TCanvas; X, Y, Width, Height: Integer;
  Src: TCanvas; XSrc, YSrc: Integer; Rop: DWORD = SRCCOPY): Boolean;
begin
  Result := BitBlt(Dest.Handle, X, Y, Width, Height, Src.Handle, XSrc, YSrc, Rop);
end;

function CreateRectRgn(X1, Y1, X2, Y2: Integer): HRGN;
begin
  Result := LCLIntf.CreateRectRgn(ScaleToNative(X1), ScaleToNative(Y1), ScaleToNative(X2),
    ScaleToNative(Y2));
end;

{$IFDEF WINDOWS}
function ScrollDC(hDC: HDC; dx: longint; dy: longint; const lprcScroll: RECT;
  const lprcClip: RECT; hrgnUpdate: HRGN; lprcUpdate: LPRECT): WINBOOL;
var
  R: RECT;
begin
  if Assigned(lprcUpdate) then begin
    R := ScaleRectToNative(lprcUpdate^);
    lprcUpdate := @R;
  end;
  Result := Windows.ScrollDC(hDC, ScaleToNative(dx), ScaleToNative(dY),
    ScaleRectToNative(lprcScroll), ScaleRectToNative(lprcClip), hrgnUpdate, lprcUpdate);
end;
{$ENDIF}

function ScrollDC(Canvas: TCanvas; dx: Longint; dy: Longint;
  const lprcScroll: TRect; const lprcClip: TRect; hrgnUpdate: Handle;
  lprcUpdate: PRect): Boolean;
begin
  {$IFDEF WINDOWS}
  Result := Windows.ScrollDC(Canvas.Handle, ScaleToNative(dx), ScaleToNative(dy),
    ScaleRectToNative(lprcScroll), ScaleRectToNative(lprcClip),
    hrgnUpdate, lprcUpdate);
  {$ENDIF}
  {$IFDEF LINUX}
  // Can't do scrolling of DC under Linux, then fallback into BitBlt.
  Result := BitBltCanvas(Canvas, lprcScroll.Left + dx, lprcScroll.Top + dy, lprcScroll.Right - lprcScroll.Left, lprcScroll.Bottom - lprcScroll.Top,
    Canvas, lprcScroll.Left, lprcScroll.Top);
  {$ENDIF}
end;

function SetWindowPos(hWnd: HWND; hWndInsertAfter: HWND; X, Y, cx, cy: Integer;
  uFlags: UINT): Boolean;
begin
  LCLIntf.SetWindowPos(hWnd, hWndInsertAfter, ScaleToNative(X), ScaleToNative(Y),
    ScaleToNative(cx), ScaleToNative(cy), uFlags);
end;

function ScaleToNative(Value: Integer): Integer;
begin
  Result := Round(Value * ScreenInfo.Dpi / 96);
end;

function ScaleToNativeDist(Base, Value: Integer): Integer;
begin
  Result := ScaleToNative(Base + Value) - ScaleToNative(Base);
end;

function ScaleFromNative(Value: Integer): Integer;
begin
  Result := Round(Value * 96 / ScreenInfo.Dpi);
end;

function ScalePointToNative(Value: TPoint): TPoint;
begin
  Result.X := ScaleToNative(Value.X);
  Result.Y := ScaleToNative(Value.Y);
end;

function ScalePointFromNative(Value: TPoint): TPoint;
begin
  Result.X := ScaleFromNative(Value.X);
  Result.Y := ScaleFromNative(Value.Y);
end;

function ScaleSizeToNative(Value: TSize): TSize;
begin
  Result.Width := ScaleToNative(Value.Width);
  Result.Height := ScaleToNative(Value.Height);
end;

function ScaleSizeFromNative(Value: TSize): TSize;
begin
  Result.Width := ScaleFromNative(Value.Width);
  Result.Height := ScaleFromNative(Value.Height);
end;

function ScaleRectToNative(Value: TRect): TRect;
begin
  Result.Left := ScaleToNative(Value.Left);
  Result.Top := ScaleToNative(Value.Top);
  Result.Right := ScaleToNative(Value.Right);
  Result.Bottom := ScaleToNative(Value.Bottom);
end;

function ScaleRectFromNative(Value: TRect): TRect;
begin
  Result.Left := ScaleFromNative(Value.Left);
  Result.Top := ScaleFromNative(Value.Top);
  Result.Right := ScaleFromNative(Value.Right);
  Result.Bottom := ScaleFromNative(Value.Bottom);
end;

function ScaleFloatToNative(Value: Double): Double;
begin
  Result := Value * ScreenInfo.Dpi / 96;
end;

function ScaleFloatFromNative(Value: Double): Double;
begin
  Result := Value * 96 / ScreenInfo.Dpi;
end;

procedure WriteLog(Text: string);
var
  F: Text;
const
  FileName = 'Log.txt';
begin
  AssignFile(F, FileName);
  if FileExists(FileName) then Append(F) else Rewrite(F);
  WriteLn(F, Text);
  CloseFile(F);
end;

function GetSystemMetrics(nIndex: Integer): Integer;
begin
  Result := ScaleFromNative(LCLIntf.GetSystemMetrics(nIndex));
end;

function BitBlt(DestDC: HDC; X, Y, Width, Height: Integer; SrcDC: HDC; XSrc,
  YSrc: Integer; Rop: DWORD = SRCCOPY): Boolean;
var
  DstWidth, DstHeight: Integer;
  SrcWidth, SrcHeight: Integer;
begin
  DstWidth := ScaleToNativeDist(X, Width);
  DstHeight := ScaleToNativeDist(Y, Height);
  SrcWidth := ScaleToNativeDist(XSrc, Width);
  SrcHeight := ScaleToNativeDist(YSrc, Height);

  if (DstWidth = SrcWidth) and (DstHeight = SrcHeight) then begin
    {$IFDEF WINDOWS}
    // On Windows LCLIntf.BitBlt is slower than direct Windows BitBlt
    Result := Windows.BitBlt(DestDC, ScaleToNative(X), ScaleToNative(Y),
      DstWidth, DstHeight, SrcDC,
      ScaleToNative(XSrc), ScaleToNative(YSrc), Rop);
    {$ELSE}
    Result := LCLIntf.BitBlt(DestDC, ScaleToNative(X), ScaleToNative(Y),
      DstWidth, DstHeight, SrcDC,
      ScaleToNative(XSrc), ScaleToNative(YSrc), Rop);
    {$ENDIF}
  end else begin
    {$IFDEF WINDOWS}
    // On Windows LCLIntf.BitBlt is slower than direct Windows BitBlt
    Result := Windows.BitBlt(DestDC, ScaleToNative(X), ScaleToNative(Y),
      Min(SrcWidth, DstWidth), Min(DstHeight, SrcHeight), SrcDC,
      ScaleToNative(XSrc), ScaleToNative(YSrc), Rop);

    // Instead calling StretchBlt for entire region try to draw missing part with BitBlt
    if DstWidth > SrcWidth then begin
      Windows.BitBlt(DestDC, ScaleToNative(X) + SrcWidth, ScaleToNative(Y),
        DstWidth - SrcWidth, DstHeight, SrcDC,
        ScaleToNative(XSrc) + SrcWidth - (DstWidth - SrcWidth), ScaleToNative(YSrc), Rop);
    end;
    if DstHeight > SrcHeight then begin
      Windows.BitBlt(DestDC, ScaleToNative(X), ScaleToNative(Y) + SrcHeight,
        DstWidth, DstHeight - SrcHeight, SrcDC,
        ScaleToNative(XSrc), ScaleToNative(YSrc) + SrcHeight - (DstHeight - SrcHeight), Rop);
    end;
    {$ELSE}
    Result := LCLIntf.BitBlt(DestDC, ScaleToNative(X), ScaleToNative(Y),
      Min(SrcWidth, DstWidth), Min(DstHeight, SrcHeight), SrcDC,
      ScaleToNative(XSrc), ScaleToNative(YSrc), Rop);

    // Instead calling StretchBlt for entire region try to draw missing part with BitBlt
    if DstWidth > SrcWidth then begin
      LCLIntf.BitBlt(DestDC, ScaleToNative(X) + SrcWidth, ScaleToNative(Y),
        DstWidth - SrcWidth, DstHeight, SrcDC,
        ScaleToNative(XSrc) + SrcWidth - (DstWidth - SrcWidth), ScaleToNative(YSrc), Rop);
    end;
    if DstHeight > SrcHeight then begin
      LCLIntf.BitBlt(DestDC, ScaleToNative(X), ScaleToNative(Y) + SrcHeight,
        DstWidth, DstHeight - SrcHeight, SrcDC,
        ScaleToNative(XSrc), ScaleToNative(YSrc) + SrcHeight - (DstHeight - SrcHeight), Rop);
    end;
    {$ENDIF}
  end;
end;

end.

